#!/bin/bash
set -eo pipefail #  exclude -u on purpose to avoid issues with param checking
#
#  SPDX-License-Identifier: LGPL-2.1-or-later
#
#  Copyright (c) 2022 Valve.
#  Maintainer: Guilherme G. Piccoli <gpiccoli@igalia.com>
#
#  This is the generic alpm-hook script that handles kernel/package
#  installation for dracut users, on behalf of kdumpst. Covers both manual
#  invocation (to create the initramfs image for a given kernel) as
#  well as kernel package installation/removal trigger. It also handles
#  the event of installing the package dracut, which requires the
#  installation of initramfs hooks to be executed early in boot time,
#  in case of a kdump.
#

#  This function has the purpose of loading the necessary external
#  variables, in the form of one (or more) configuration file(s). If the
#  procedure fails, we must abort - otherwise it'll fail in a later stage.
load_kdumpst_config() {
	HAVE_CFG_FILES=0
	shopt -s nullglob
	for cfg in "/usr/share/kdumpst.d"/*; do
		if [ -f "$cfg" ]; then
			. "$cfg"
			HAVE_CFG_FILES=1
		fi
	done

	for cfg in "/etc/kdumpst.d"/*; do
		if [ -f "$cfg" ]; then
			. "$cfg"
			HAVE_CFG_FILES=1
		fi
	done
	shopt -u nullglob

	if [ ${HAVE_CFG_FILES} -eq 0 ]; then
		logger "kdumpst: no config files in /usr/share/kdumpst.d/ or /etc/kdumpst.d/ - aborting."
		exit 1
	fi
}

#  Functions to deal with dracut specifics, both for building
#  the initramfs for dracut users, but also with regards to
#  installing dracut specific hooks/scripts.
#
#  IMPORTANT: it is assumed that kdumpst configuration was loaded
#  before running any of these functions!
#
create_initramfs_dracut() {
	rm -f "${MOUNT_FOLDER}/kdump-initrd-$1.img"

	COMPRESS="--compress=gzip"
	MAJOR="$(echo "$1" | cut -f1 -d\.)"
	MINOR="$(echo "$1" | cut -f2 -d\.)"

	#  Zstd is FAST, but only supported in kernels 5.9+.
	if [ "${MAJOR}" -gt 5 ] || ([ "${MAJOR}" -eq 5 ] && [ "${MINOR}" -ge 9 ]); then
		COMPRESS="--compress=zstd"
	fi

	DRACUT_NO_XATTR=1 dracut "${COMPRESS}" --no-early-microcode --host-only -q -m\
	"bash systemd systemd-initrd systemd-sysusers modsign dbus-daemon kdump dbus udev-rules dracut-systemd base fs-lib shutdown"\
	--kver "$1" "${MOUNT_FOLDER}/kdump-initrd-$1.img"

	if [ -s "${MOUNT_FOLDER}/kdump-initrd-$1.img" ]; then
		logger "kdumpst: created dracut minimal initramfs"
	fi
}

dracut_installation() {
	HOOKS_DIR="/usr/lib/kdumpst/dracut/"

	DRACUT_DIR="$(pkg-config --variable=dracutmodulesdir dracut 2>/dev/null)"
	if [ -z "${DRACUT_DIR}" ]; then
		DRACUT_DIR="/usr/lib/dracut/modules.d/"
	fi

	if [ ! -d "${DRACUT_DIR}"/55kdump/ ]; then
		install -D -m0755 "${HOOKS_DIR}"/kdump-collect.sh "${DRACUT_DIR}"/55kdump/kdump-collect.sh
		install -D -m0755 "${HOOKS_DIR}"/module-setup.sh "${DRACUT_DIR}"/55kdump/module-setup.sh
		logger "kdumpst: dracut hooks/scripts installed"
	fi
}

load_kdumpst_config

#  General comment of a caveat here: dracut_installation() must be
#  reentrant and only install the dracut handlers if they're not
#  already installed. This is necessary due to a kinda chicken-egg
#  problem: the first time initramfs creation is attempted, we did't
#  necessarily run the dracut package installation hooks yet, hence
#  we may be unable to properly create the initramfs image. "Easy"
#  solution is to just try to install them always, it's cheap and not
#  dependent of package install time magic.

#  Parameter passing case - since this is invoked by the kdump-load
#  script, we must be sure that the dracut package is available.
if [ -n "$1" ]; then

	if command -v dracut 1>/dev/null; then
		dracut_installation

        #  Worth noticing that it is possible (though unlikely) that during
        #  the kdumpst package installation, users are running a previously
        #  uninstalled kernel. Imagine if the user upgrades the kernel, and
        #  then install kdumpst before rebooting. With that, creating an
        #  initrd fails here since it relies on "uname -r". To prevent such
        #  unhandled error, check if this kernel's modules directory exists.

        if [ ! -d "/lib/modules/$1" ]; then
            logger "kdumpst: no kernel modules dir; defer initrd creation to next boot"
            exit 0
        fi

		create_initramfs_dracut "$1"
	fi
	exit 0
fi

while read -r line; do
	#  First case is the dracut package installation.
	if [[ "$line" != */vmlinuz ]]; then
		dracut_installation
		exit 0
	fi

	#  If reaching this point, we're installing/removing the kernel image.
	VERSION="$(basename "$(dirname "$line")")"

	#  If the file exists, means it's an installation step;
	#  notice that alpm suppresses the leading '/'.
	if [ -f /"$line" ]; then
		dracut_installation
		create_initramfs_dracut "$VERSION"
	else
		rm -f "${MOUNT_FOLDER}"/kdump-initrd-"$VERSION".img

	fi
	exit 0
done
