#!/bin/bash
set -eo pipefail #  exclude -u on purpose to avoid issues with param checking
#
#  SPDX-License-Identifier: LGPL-2.1-or-later
#
#  Copyright (c) 2022 Valve.
#  Maintainer: Guilherme G. Piccoli <gpiccoli@igalia.com>
#
#  This is the generic alpm-hook script that handles kernel/package
#  installation for mkinitcpio users, on behalf of kdumpst. Covers both manual
#  invocation (to create the initramfs image for a given kernel) as
#  well as kernel package installation/removal trigger. It also handles
#  the event of installing the package mkinitcpio, which requires the
#  installation of initramfs hooks to be executed early in boot time,
#  in case of a kdump.
#

#  This function has the purpose of loading the necessary external
#  variables, in the form of one (or more) configuration file(s). If the
#  procedure fails, we must abort - otherwise it'll fail in a later stage.
load_kdumpst_config() {
	HAVE_CFG_FILES=0
	shopt -s nullglob
	for cfg in "/usr/share/kdumpst.d"/*; do
		if [ -f "$cfg" ]; then
			. "$cfg"
			HAVE_CFG_FILES=1
		fi
	done

	for cfg in "/etc/kdumpst.d"/*; do
		if [ -f "$cfg" ]; then
			. "$cfg"
			HAVE_CFG_FILES=1
		fi
	done
	shopt -u nullglob

	if [ ${HAVE_CFG_FILES} -eq 0 ]; then
		logger "kdumpst: no config files in /usr/share/kdumpst.d/ or /etc/kdumpst.d/ - aborting."
		exit 1
	fi
}

#  Functions to deal with initcpio specifics, both for building
#  the initramfs for mkinitcpio users, but also with regards to
#  installing its specific hooks.
#
#  IMPORTANT: it is assumed that kdumpst configuration was loaded
#  before running any of these functions!
#
create_initramfs_mkinitcpio() {
	rm -f "${MOUNT_FOLDER}/kdump-initrd-$1.img"

	mkinitcpio -A kdump -g "${MOUNT_FOLDER}/kdump-initrd-$1.img" -k "$1" 1>/dev/null

	if [ -s "${MOUNT_FOLDER}/kdump-initrd-$1.img" ]; then
		logger "kdumpst: created initcpio minimal initramfs"
	fi
}

mkinitcpio_installation() {
	KDUMP_HOOKS_DIR="/usr/lib/kdumpst/initcpio/"
	INITCPIO_HOOKS="/usr/lib/initcpio/hooks"
	INITCPIO_INST="/usr/lib/initcpio/install"

	if [ ! -e "${INITCPIO_HOOKS}"/kdump ] || [ ! -e "${INITCPIO_INST}"/kdump ]; then
		install -D -m0644 "${KDUMP_HOOKS_DIR}"/kdump.hook "${INITCPIO_HOOKS}"/kdump
		install -D -m0644 "${KDUMP_HOOKS_DIR}"/kdump.install "${INITCPIO_INST}"/kdump
		logger "kdumpst: initcpio hooks installed"
	fi
}

load_kdumpst_config

#  General comment of a caveat here: mkinitcpio_installation() must be
#  reentrant and only install the mkinitcpio handlers if they're not
#  already installed. This is necessary due to a kinda chicken-egg
#  problem: the first time initramfs creation is attempted, we did't
#  necessarily run the mkinitcpio package installation hooks yet, hence
#  we may be unable to properly create the initramfs image. "Easy"
#  solution is to just try to install them always, it's cheap and not
#  dependent of package install time magic.

#  Parameter passing case - since this is invoked by the kdump-load
#  script, we must be sure that the mkinitcpio package is available.
if [ -n "$1" ]; then

	if command -v mkinitcpio 1>/dev/null; then
		mkinitcpio_installation

        #  Worth noticing that it is possible (though unlikely) that during
        #  the kdumpst package installation, users are running a previously
        #  uninstalled kernel. Imagine if the user upgrades the kernel, and
        #  then install kdumpst before rebooting. With that, creating an
        #  initrd fails here since it relies on "uname -r". To prevent such
        #  unhandled error, check if this kernel's modules directory exists.

        if [ ! -d "/lib/modules/$1" ]; then
            logger "kdumpst: no kernel modules dir; defer initrd creation to next boot"
            exit 0
        fi

		create_initramfs_mkinitcpio "$1"
	fi
	exit 0
fi

while read -r line; do
	#  First case is the mkinitcpio package installation.
	if [[ "$line" != */vmlinuz ]]; then
		mkinitcpio_installation
		exit 0
	fi

	#  If reaching this point, we're installing/removing the kernel image.
	VERSION="$(basename "$(dirname "$line")")"

	#  If the file exists, means it's an installation step;
	#  notice that alpm suppresses the leading '/'.
	if [ -f /"$line" ]; then
		mkinitcpio_installation
		create_initramfs_mkinitcpio "$VERSION"
	else
		rm -f "${MOUNT_FOLDER}"/kdump-initrd-"$VERSION".img

	fi
	exit 0
done
